from abc import abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, List
import torch
from ml_gym.batching.batch import DatasetBatch, EvaluationBatchResult, InferenceResultBatch
from ml_gym.data_handling.dataset_loader import DatasetLoader
from ml_gym.gym.inference_component import InferenceComponent
from ml_gym.gym.stateful_components import StatefulComponent
from ml_gym.metrics.metrics import Metric
from ml_gym.models.nn.net import NNModel
from ml_gym.loss_functions.loss_functions import Loss, LossWarmupMixin
import tqdm
from ml_gym.util.logger import LogLevel, ConsoleLogger
import numpy as np


class AbstractEvaluator(StatefulComponent):
    @abstractmethod
    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        raise NotImplementedError

    # @abstractmethod
    # def warm_up(self, model: NNModel, device: torch.device):
    #     raise NotImplementedError


class Evaluator(AbstractEvaluator):
    def __init__(self, eval_component: 'EvalComponentIF'):
        self.eval_component = eval_component

    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        return self.eval_component.evaluate(model, device)

    # def warm_up(self, model: NNModel, device: torch.device):
    #     self.eval_component.warm_up(model, device)


class EvalComponentIF(StatefulComponent):
    # @abstractmethod
    # def warm_up(self, model: NNModel, device: torch.device):
    #     raise NotImplementedError

    @abstractmethod
    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        raise NotImplementedError


class EvalComponent(EvalComponentIF):
    """This thing always comes with batteries included, i.e., datasets, loss functions etc. are all already stored in here."""

    def __init__(self, inference_component: InferenceComponent, metrics: List[Metric],
                 loss_funs: Dict[str, Loss], dataset_loaders: Dict[str, DatasetLoader], train_split_name: str, show_progress: bool = False,
                 cpu_target_subscription_keys: List[str] = None, cpu_prediction_subscription_keys: List[str] = None):
        self.loss_funs = loss_funs
        self.inference_component = inference_component
        self.metrics = metrics
        self.dataset_loaders = dataset_loaders
        self.train_split_name = train_split_name
        self.show_progress = show_progress
        self.cpu_target_subscription_keys = set(cpu_target_subscription_keys)
        self.cpu_prediction_subscription_keys = set(cpu_prediction_subscription_keys)
        self.logger = ConsoleLogger("logger_eval_component")

    # def warm_up(self, model: NNModel, device: torch.device):
    #     def init_loss_funs(batch: InferenceResultBatch):
    #         for _, loss_fun in self.loss_funs.items():
    #             if isinstance(loss_fun, LossWarmupMixin):
    #                 loss_fun.warm_up(batch)
    #                 loss_fun.finish_warmup()

    #     if any([isinstance(loss_fun, LossWarmupMixin) for _, loss_fun in self.loss_funs.items()]):
    #         self.logger.log(LogLevel.INFO, "Running warmup...")
    #         prediction_batches = self.map_batches(fun=self.forward_batch,
    #                                               fun_params={"calculation_device": device, "model": model,
    #                                                           "result_device": torch.device("cpu")},
    #                                               loader=self.dataset_loaders[self.train_split_name],
    #                                               show_progress=self.show_progress)
    #         prediction_batch = InferenceResultBatch.combine(prediction_batches)
    #         init_loss_funs(prediction_batch)
    #     else:
    #         self.logger.log(LogLevel.INFO, "Skipping evaluation warmup. No special loss functions to be initialized.")

    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        return [self.evaluate_dataset_split(model, device, split_name, loader) for split_name, loader in self.dataset_loaders.items()]

    def evaluate_dataset_split(self, model: NNModel, device: torch.device, split_name: str, dataset_loader: DatasetLoader) -> EvaluationBatchResult:
        if self.show_progress:
            dataset_loader = tqdm.tqdm(dataset_loader, desc="Batches processed:")

        batch_losses = []
        inference_result_batches_cpu = []
        for batch in dataset_loader:
            inference_result_batch = self.forward_batch(dataset_batch=batch, model=model, device=device)
            batch_loss = self._calculate_loss_scores(inference_result_batch)
            batch_losses.append(batch_loss)
            irb_filtered = inference_result_batch.split_results(predictions_keys=self.cpu_prediction_subscription_keys,
                                                                target_keys=self.cpu_target_subscription_keys,
                                                                device=torch.device("cpu"))
            inference_result_batches_cpu.append(irb_filtered)

        # calc metrics
        prediction_batch = InferenceResultBatch.combine(inference_result_batches_cpu)
        metric_scores = self._calculate_metric_scores(prediction_batch)

        # calc losses
        loss_keys = batch_losses[0].keys()
        loss_scores = {key: [np.mean([l[key] for l in batch_losses])] for key in loss_keys}

        evaluation_result = EvaluationBatchResult(losses=loss_scores,
                                                  metrics=metric_scores,
                                                  dataset_name=dataset_loader.dataset_name,
                                                  split_name=split_name)
        return evaluation_result

    def _get_metric_fun(self, identifier: str, target_subscription: Enum, prediction_subscription: Enum,
                        metric_fun: Callable, params: Dict[str, Any]) -> Metric:
        return Metric(identifier, target_subscription, prediction_subscription, metric_fun, params)

    def forward_batch(self, dataset_batch: DatasetBatch, model: NNModel, device: torch.device) -> InferenceResultBatch:
        model = model.to(device)
        dataset_batch.to_device(device)
        inference_result_batch = self.inference_component.predict(model, dataset_batch)
        return inference_result_batch

    def _calculate_metric_scores(self, inference_batch: InferenceResultBatch) -> Dict[str, List[float]]:
        return {metric.tag: [metric(inference_batch)] for metric in self.metrics}

    def _calculate_loss_scores(self, forward_batch: InferenceResultBatch) -> Dict[str, List[float]]:
        return {loss_key: self._get_batch_loss(loss_fun, forward_batch) for loss_key, loss_fun in self.loss_funs.items()}

    def _get_batch_loss(self, loss_fun: Loss, forward_batch: InferenceResultBatch) -> List[float]:
        loss = loss_fun(forward_batch)
        loss = [loss.detach().item()]
        return loss



import logging
import os
from multiprocessing import Queue, Process
from dataclasses import dataclass
import time
from abc import ABC, abstractmethod
from enum import Enum
import sys
import traceback
from ml_gym.error_handling.exception import SingletonAlreadyInstantiatedError


class LogLevel(Enum):
    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10
    NOTSET = 0


@dataclass
class Message:
    logger_id: str
    message_string: str
    unix_timestamp: float
    level: LogLevel


class MLgymLoggerIF(ABC):

    def __init__(self, logger_id: str):
        self.logger_id = logger_id

    def build_message(self, level: LogLevel, message) -> Message:
        timestamp = time.time()
        message = Message(message_string=message, unix_timestamp=timestamp, level=level, logger_id=self.logger_id)
        return message

    @abstractmethod
    def log(self, level: LogLevel, message: str):
        raise NotImplementedError


class ConsoleLogger(MLgymLoggerIF):

    def __init__(self, logger_id: str):
        super().__init__(logger_id)
        self.logger = ConsoleLogger._get_console_logger(logger_id)

    def log(self, level: LogLevel, message: str):
        self.logger.log(level.value, message)

    @staticmethod
    def _get_console_logger(logger_id: str):
        # create / get logger
        logger = logging.getLogger(logger_id)
        # if logger has not been created
        if not logger.hasHandlers():
            logger.setLevel(logging.DEBUG)

            # create console handler with a higher log level
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            # create formatter and add it to the handlers
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            # add the handlers to the logger
            logger.addHandler(ch)
        return logger


class QLogger(MLgymLoggerIF):

    def __init__(self, logger_id: str, queue: Queue):
        super().__init__(logger_id)
        self.queue = queue

    def log(self, level: LogLevel, message: str):
        message = self.build_message(level, message)
        self.queue.put(message)


class QueuedLogging:
    _instance: "QueuedLogging" = None

    @classmethod
    def start_logging(cls, log_msg_queue: Queue, log_dir_path: str):
        if cls._instance is None:
            cls._instance = QueuedLogging(log_msg_queue, log_dir_path)
            cls._instance._start_listening()
        else:
            raise SingletonAlreadyInstantiatedError()

    @staticmethod
    def get_qlogger(logger_id: str) -> QLogger:
        return QLogger(logger_id, QueuedLogging._instance.log_msg_queue)

    def __init__(self, log_msg_queue: Queue, log_dir_path: str):
        if self._instance is not None:
            raise SingletonAlreadyInstantiatedError()
        self.log_msg_queue = log_msg_queue
        self.log_dir_path = log_dir_path
        self.listener_process = None

    def _start_listening(self):
        self.listener_process = Process(target=QueuedLogging._listener_process, args=(self.log_msg_queue, self.log_dir_path))
        self.listener_process.start()

    @staticmethod
    def stop_listener():
        QueuedLogging._instance.log_msg_queue.put_nowait(None)

    @staticmethod
    def _get_logger(log_dir_path: str, logger_id: str):
        logger = logging.getLogger(logger_id)
        log_file_path = os.path.join(log_dir_path, f"{logger_id}.log")
        os.makedirs(os.path.dirname(log_file_path), exist_ok=True)
        logger.setLevel(logging.DEBUG)
        # create file handler which logs even debug messages
        fh = logging.FileHandler(log_file_path)
        fh.setLevel(logging.DEBUG)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(fh)
        logger.addHandler(ch)
        return logger

    @staticmethod
    def _listener_process(queue: Queue, log_dir_path: str):
        loggers = {}
        while True:
            try:
                message: Message = queue.get()
                if message is None:  # We send this as a sentinel to tell the listener to quit.
                    break
                if message.logger_id not in loggers:
                    loggers[message.logger_id] = QueuedLogging._get_logger(log_dir_path, message.logger_id)
                loggers[message.logger_id].log(level=message.level.value, msg=message.message_string)
            except Exception:
                print('Whoops! Problem:', file=sys.stderr)
                traceback.print_exc(file=sys.stderr)


def get_console_logger(name: str):
    # create logger with 'spam_application'
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)
    return logger



from abc import abstractmethod
from ml_gym.models.nn.net import NNModel
from dashify.logging.dashify_logging import ExperimentInfo
from ml_gym.gym.evaluator import Evaluator
from ml_gym.gym.trainer import Trainer
import torch
from ml_gym.gym.stateful_components import StatefulComponent
from ml_gym.persistency.io import DashifyReader
from ml_gym.persistency.io import DashifyWriter
from enum import Enum
from typing import List
from torch.optim.optimizer import Optimizer
from ml_gym.util.logger import LogLevel, ConsoleLogger


class AbstractGymJob(StatefulComponent):
    class Mode(Enum):
        EVAL = "eval"
        TRAIN = "training"

    def __init__(self, experiment_info: ExperimentInfo):
        super().__init__()
        self._experiment_info = experiment_info
        # TODO: This is only a console logger that does not store on disk.
        # Since GymJob is a StatefulComponent we would have to implement serialization
        # for the queue in the QLogger. Solving this issue is left for the future
        self.logger = ConsoleLogger("logger_gym_job")

    @property
    def experiment_info(self) -> ExperimentInfo:
        return self._experiment_info

    @abstractmethod
    def execute(self, device: torch.device):
        raise NotImplementedError

    @staticmethod
    def from_blue_print(blue_print) -> 'AbstractGymJob':
        return blue_print.construct()

    def save_state_of_stateful_components(self, measurement_id: int):
        state = self.get_state()
        DashifyWriter.save_state(experiment_info=self.experiment_info, data_dict=state, measurement_id=measurement_id)

    def restore_state_in_stateful_components(self, measurement_id: int):
        state = DashifyReader.load_state(experiment_info=self.experiment_info, measurement_id=measurement_id)
        self.set_state(state)


class GymJob(AbstractGymJob):

    def __init__(self, mode: AbstractGymJob.Mode.TRAIN, model: NNModel, optimizer, trainer: Trainer,
                 evaluator: Evaluator, experiment_info: ExperimentInfo, epochs: List[int]):
        super().__init__(experiment_info)
        self.mode = mode
        self.model = model
        self.optimizer_partial = optimizer  # TODO this is a little bit ugly.
        self.epochs = epochs
        self.evaluator = evaluator
        self.trainer = trainer
        if self.mode == AbstractGymJob.Mode.TRAIN and epochs[0] == 0:
            DashifyWriter.save_model_state(model.state_dict(), self._experiment_info, 0)
            optimizer: Optimizer = self.optimizer_partial(params=self.model.parameters())
            DashifyWriter.save_optimizer_state(optimizer.state_dict(), self._experiment_info, 0)
        self._execution_method = self._execute_train if mode == GymJob.Mode.TRAIN else self._execute_eval

    def _train_step(self, device: torch.device, measurement_id: int) -> NNModel:
        self.restore_state_in_stateful_components(measurement_id - 1)
        # load model and optimizer
        model_state = DashifyReader.load_model_state(self._experiment_info, measurement_id - 1)
        self.model.load_state_dict(model_state)
        self.model.to(device)
        optimizer: Optimizer = self.optimizer_partial(params=self.model.parameters())
        optimizer_state = DashifyReader.load_optimizer_state(self._experiment_info, measurement_id - 1)
        optimizer.load_state_dict(optimizer_state)
        # train
        model = self.trainer.train_epoch(self.model, optimizer, device)
        # save model and optimizer
        DashifyWriter.save_model_state(model.state_dict(), self._experiment_info, measurement_id)
        DashifyWriter.save_optimizer_state(optimizer.state_dict(), self._experiment_info, measurement_id)
        self.save_state_of_stateful_components(measurement_id=measurement_id)
        return model

    def _evaluation_step(self, device: torch.device, measurement_id: int):
        self.restore_state_in_stateful_components(measurement_id)
        model_state = DashifyReader.load_model_state(self._experiment_info, measurement_id)
        self.model.load_state_dict(model_state)
        self.model.to(device)
        evaluation_result = self.evaluator.evaluate(self.model, device)
        DashifyWriter.log_measurement_result(evaluation_result, self._experiment_info, measurement_id=measurement_id)

    def _warmup(self, device: torch.device, measurement_id: int):
        # self.evaluator.warm_up(self.model, device)
        # self.trainer.warm_up(self.model, device)
        self.save_state_of_stateful_components(measurement_id=measurement_id)

    def execute(self, device: torch.device):
        """ Executes the job

        Args:
            device: torch device either CPUs or a specified GPU
        """
        self._execution_method(device)

    def _execute_train(self, device: torch.device):
        step_epochs = self.epochs[1:]
        if len(self.epochs) > 0 and self.epochs[0] == 0:
            # perform warmup for trainer and evaluator components
            # self.logger.log(LogLevel.DEBUG, "Executing warmup step")
            self._warmup(device, measurement_id=0)
            # get initial results
            # self.logger.log(LogLevel.DEBUG, "Executing initial evaluation step")
            self._evaluation_step(device, measurement_id=0)
            # train and evaluate

        for epoch in step_epochs:
            # self.logger.log(LogLevel.DEBUG, "Executing training step")
            self._train_step(device, measurement_id=epoch)
            # self.logger.log(LogLevel.DEBUG, "Executing evaluation step")
            self._evaluation_step(device, epoch)

    def _execute_eval(self, device: torch.device):
        for epoch in self.epochs:
            self._evaluation_step(device, measurement_id=epoch)



from abc import ABC, abstractmethod
from typing import Dict, Any


class ValidatorIF(ABC):

    @abstractmethod
    def run(self, gs_config: Dict[str, Any], num_epochs: int):
        raise NotImplementedError



import copy
from typing import Dict, Any, List, Optional, Type, Union
from collections import namedtuple
from dataclasses import dataclass, field
from ml_gym.error_handling.exception import ComponentConstructionError
from ml_gym.blueprints.constructables import ComponentConstructable, DatasetIteratorConstructable, \
    DatasetIteratorSplitsConstructable, Requirement, DataLoadersConstructable, DatasetRepositoryConstructable, \
    OptimizerConstructable, ModelRegistryConstructable, ModelConstructable, LossFunctionRegistryConstructable, \
    MetricFunctionRegistryConstructable, TrainerConstructable, EvaluatorConstructable, MappedLabelsIteratorConstructable, \
    FilteredLabelsIteratorConstructable, FeatureEncodedIteratorConstructable, CombinedDatasetIteratorConstructable, \
    DataCollatorConstructable, PredictionPostProcessingRegistryConstructable, TrainComponentConstructable, EvalComponentConstructable, \
    IteratorViewConstructable, OneHotEncodedTargetsIteratorConstructable, InMemoryDatasetIteratorConstructable
from ml_gym.error_handling.exception import InjectMappingNotFoundError
# from ml_gym.util.logger import LogLevel, ConsoleLogger


class Injector:

    def __init__(self, mapping: Dict[str, Any], raise_mapping_not_found: bool = False):
        self.mapping = mapping
        self.raise_mapping_not_found = raise_mapping_not_found

    def inject_pass(self, component_parameters: Dict) -> Dict[str, Any]:
        def inject(tree: Union[Dict, List]) -> Dict[str, Any]:
            if isinstance(tree, dict):
                for key, sub_tree in tree.items():
                    if key == "injectable":
                        if tree["injectable"]["id"] not in self.mapping:
                            if self.raise_mapping_not_found:
                                raise InjectMappingNotFoundError
                            else:
                                tree[key] = sub_tree
                        else:
                            tree = self.mapping[sub_tree["id"]]
                    else:
                        tree[key] = inject(sub_tree)
                return tree
            elif isinstance(tree, list):
                return [inject(sub_tree) for sub_tree in tree]
            else:
                return tree
        copied_dict = copy.deepcopy(component_parameters)
        injected = {key: inject(parameter) for key, parameter in copied_dict.items()}
        return injected


@dataclass
class ComponentRepresentation:
    """ Direct class representation of components in the config dictionary
    """
    name: str
    component_type_key: str
    variant_key: str
    config: Dict = field(default_factory=dict)
    requirements: Dict[str, "RequirementRepresentation"] = field(default_factory=dict)

    def __str__(self):
        return f"ComponentRepresentation(name='{self.name}'," \
            f"component_type_key='{self.component_type_key}'," \
            f"variant_key='{self.variant_key}'," \
            f"config={str(self.config)}," \
            f"hash='{self.hash}'," \
            f"requirements={self.requirements}])"

    def __repr__(self):
        return self.__str__()


@dataclass
class RequirementRepresentation:
    """ Class representation of a requirement of a component defined in the config.
    """
    name: str
    component_name: str
    subscription: List[int]


class ComponentFactory:
    """
    The component factory builds the machine learning components like dataset iterators, dataset splitters, models, trainers etc.
    Each component type can have multiple variants. This relationship is captured by component_factory_registry and
    ComponentVariantsRegistry. The ComponentFactoryRegistry maps a component_key to a ComponentVariantsRegistry.
    The ComponentVariantsRegistry maps a variant key to the concrete ComponentConstructable.
    """

    class ComponentVariantsRegistry:
        def __init__(self):
            self.constructables: Dict[str, ComponentConstructable] = {}

        def construct(self, variant_key: str, component_name: str, config: Dict[str, Any] = None, requirements: List[Any] = None):
            if config is None:
                config = {}

            try:
                constructable = self.constructables[variant_key]
                component = constructable(component_identifier=component_name, requirements=requirements, **config).construct()
            except Exception as e:
                raise ComponentConstructionError(f"Error during component creation from {constructable}") from e
            return component

        def register_variant(self, variant_key: str, component_constructable_type: Type[ComponentConstructable]):
            self.constructables[variant_key] = component_constructable_type

    def __init__(self, injector: Injector = None):
        self.injector = injector
        ComponentVariant = namedtuple('ComponentVariant', ['component_key',
                                                           'variant_key',
                                                           'component_constructable_type'])
        # register the default components with their variants
        default_component_variants = [
            ComponentVariant("DATASET_REPOSITORY", "DEFAULT", DatasetRepositoryConstructable),
            ComponentVariant("DATASET_ITERATORS", "DEFAULT", DatasetIteratorConstructable),
            ComponentVariant("SPLITTED_DATASET_ITERATORS", "RANDOM", DatasetIteratorSplitsConstructable),
            ComponentVariant("COMBINED_DATASET_ITERATORS", "DEFAULT", CombinedDatasetIteratorConstructable),
            ComponentVariant("IN_MEMORY_DATASET_ITERATORS", "DEFAULT", InMemoryDatasetIteratorConstructable),
            ComponentVariant("FILTERED_LABELS_ITERATOR", "DEFAULT", FilteredLabelsIteratorConstructable),
            ComponentVariant("ONE_HOT_ENCODED_TARGETS_ITERATOR", "DEFAULT", OneHotEncodedTargetsIteratorConstructable),
            ComponentVariant("ITERATOR_VIEW", "DEFAULT", IteratorViewConstructable),
            ComponentVariant("MAPPED_LABELS_ITERATOR", "DEFAULT", MappedLabelsIteratorConstructable),
            ComponentVariant("DATA_COLLATOR", "DEFAULT", DataCollatorConstructable),
            ComponentVariant("FEATURE_ENCODED_ITERATORS", "DEFAULT", FeatureEncodedIteratorConstructable),
            ComponentVariant("DATA_LOADER", "DEFAULT", DataLoadersConstructable),
            ComponentVariant("OPTIMIZER", "DEFAULT", OptimizerConstructable),
            ComponentVariant("MODEL_REGISTRY", "DEFAULT", ModelRegistryConstructable),
            ComponentVariant("LOSS_FUNCTION_REGISTRY", "DEFAULT", LossFunctionRegistryConstructable),
            ComponentVariant("METRIC_REGISTRY", "DEFAULT", MetricFunctionRegistryConstructable),
            ComponentVariant("PREDICTION_POSTPROCESSING_REGISTRY", "DEFAULT", PredictionPostProcessingRegistryConstructable),
            ComponentVariant("MODEL", "DEFAULT", ModelConstructable),
            ComponentVariant("TRAIN_COMPONENT", "DEFAULT", TrainComponentConstructable),
            ComponentVariant("TRAINER", "DEFAULT", TrainerConstructable),
            ComponentVariant("EVAL_COMPONENT", "DEFAULT", EvalComponentConstructable),
            ComponentVariant("EVALUATOR", "DEFAULT", EvaluatorConstructable)
        ]
        self.component_factory_registry: Dict[str, Any] = {}
        for variant in default_component_variants:
            self.register_component_type(*variant)

    def register_component_type(self, component_key: str, variant_key: str, component_constructable_type: Type[ComponentConstructable]):
        if component_key not in self.component_factory_registry:
            component_variants_registry = ComponentFactory.ComponentVariantsRegistry()
            self.component_factory_registry[component_key] = component_variants_registry
        self.component_factory_registry[component_key].register_variant(variant_key, component_constructable_type)

    def _calc_dependency_graph(self, component_config: Dict) -> Dict[str, ComponentRepresentation]:
        """Returns a dict that maps component hash values to components.
        Each component can have requirements which in return are components again.
        Thereby, defining the dependency relationship of components, e.g., trainer depends on model."

        Args:
            component_config (Dict): Raw config describing the components and their dependencies

        Returns:
            Dict[str, ComponentRepresentation]: [description]
        """
        def create_component_representation(component_config: Dict, component_name: str = None) -> ComponentRepresentation:
            requirements = {}
            if "requirements" in component_config:
                requirements = {}
                for requirement in component_config.pop("requirements"):
                    requirement_component_name = requirement["component_name"]
                    name = requirement["name"]
                    subscription = requirement.get("subscription")
                    req_rep = RequirementRepresentation(name=name, component_name=requirement_component_name, subscription=subscription)
                    requirements[name] = req_rep
            # inject components
            if self.injector and "config" in component_config:
                component_config["config"] = self.injector.inject_pass(component_config["config"])
            try:
                component_representation = ComponentRepresentation(name=component_name, requirements=requirements, **component_config)
                return component_representation
            except Exception as e:
                raise ComponentConstructionError(
                    f"Error during component representation instantiation. component_name: {component_name}") from e

        component_config = copy.deepcopy(component_config)  # needed because the pop call in the auxilary function is destructive
        component_representations = {}
        for component_name, component_dict in component_config.items():
            component_representation = create_component_representation(component_config=component_dict, component_name=component_name)
            component_representations[component_representation.name] = component_representation
        return component_representations

    def build_components_from_config(self, component_config: Dict, names_of_components_to_construct: List[str]) -> Dict:
        """Builds the components and returns a mapping from component name to component.
        Note, that dependencies are always rebuilt and not reused!

        Args:
            component_config (Dict): Raw config describing the components and their dependencies

        Returns:
            Dict: Component dictionary (component name -> Component)
        """
        def build_component(component_name: str, component_representation_graph: Dict[str, ComponentRepresentation], components: Dict[str, Any]):
            component_representation = component_representation_graph[component_name]
            # build the requirements
            requirement_components = {name: build_component(requirement.component_name, component_representation_graph, {})
                                      for name, requirement in component_representation.requirements.items()}

            # collect the requirements
            requirements = {name: Requirement(requirement_components[name], requirement.subscription)
                            for name, requirement in component_representation.requirements.items()}
            # build the requested component
            component_variants_registry = self.component_factory_registry[component_representation.component_type_key]
            component = component_variants_registry.construct(
                component_representation.variant_key, component_representation.name, component_representation.config, requirements)
            return component

        # calculate the dependency graph of components
        component_representation_graph = self._calc_dependency_graph(component_config)
        # build each component
        components: Dict[str, Any] = {}  # maps
        components = {component_name: build_component(component_name, component_representation_graph, {})
                      for component_name, component_representation in component_representation_graph.items()
                      if component_name in names_of_components_to_construct}
        return components



from abc import abstractmethod
import glob
import os
from abc import ABC
from typing import Callable, Dict, List, Any
import json
import tqdm
from collections import defaultdict
import numpy as np


class EvaluationAnalyzerIF(ABC):

    @abstractmethod
    def analyze(self):
        raise NotImplementedError


class ExperimentRatingStrategyIf(ABC):

    @abstractmethod
    def __call__(self) -> float:
        raise NotImplementedError


class ExperimentRepresentation:
    def __init__(self, config_path: str, metric_path: str):
        self.config_path = config_path
        self.metric_path = metric_path
        with open(config_path, "r") as f:
            self.config = json.load(f)
        with open(metric_path, "r") as f:
            self.results = json.load(f)
            self.results = {key: np.array(result_list) for key, result_list in self.results.items()}

        self.experiment_id = self.config["nested_cv_experiment_information"]["config"]["experiment_id"]
        self.hyper_paramater_combination_id = self.config["nested_cv_experiment_information"]["config"]["hyper_paramater_combination_id"]
        self.outer_test_fold_id = self.config["nested_cv_experiment_information"]["config"]["outer_test_fold_id"]
        self.inner_test_fold_id = self.config["nested_cv_experiment_information"]["config"]["inner_test_fold_id"]


class NestedCVRepresentation:
    def __init__(self) -> None:
        # outer fold id -> hpc_id -> Experiment
        self.outer_experiments: Dict[int, Dict[int, ExperimentRepresentation]] = {}
        # outer fold id -> inner fold id -> hpc_id -> Experiment
        self.inner_experiments: Dict[int, Dict[int, Dict[int, ExperimentRepresentation]]] = {}

    def add_experiment(self, e: ExperimentRepresentation):
        # outer folds
        if e.inner_test_fold_id == -1:
            if e.outer_test_fold_id not in self.outer_experiments:
                self.outer_experiments[e.outer_test_fold_id] = {}
            self.outer_experiments[e.outer_test_fold_id][e.hyper_paramater_combination_id] = e
        # inner folds
        if e.inner_test_fold_id > -1:
            if e.outer_test_fold_id not in self.inner_experiments:
                self.inner_experiments[e.outer_test_fold_id] = {}
            if e.inner_test_fold_id not in self.inner_experiments[e.outer_test_fold_id]:
                self.inner_experiments[e.outer_test_fold_id][e.inner_test_fold_id] = {}
            self.inner_experiments[e.outer_test_fold_id][e.inner_test_fold_id][e.hyper_paramater_combination_id] = e

    def get_hyper_parameter_combination_ids(self) -> List[int]:
        return list(self.outer_experiments[0].keys())

    def verify(self) -> bool:
        experiment_ids = set(self.outer_experiments[0].keys())
        for _, fold in self.outer_experiments.items():
            if set(fold.keys()) != experiment_ids:
                return False

        for _, outer_fold in self.inner_experiments.items():
            for _, inner_fold in outer_fold.items():
                if set(inner_fold.keys()) != experiment_ids:
                    return False
        return True


class NestedCVReport:
    def __init__(self, nested_cv_path: str, experiments: List[ExperimentRepresentation], epoch: int):
        self.nested_cv_path = nested_cv_path
        self.experiments = experiments
        self.report = self.calc_report(self.experiments)
        self.epoch = epoch

    @staticmethod
    def calc_report(experiments: List[ExperimentRepresentation]) -> Dict[str, Any]:
        results_list: List[Dict[str, Any]] = [e.results for e in experiments]
        aggregated_metrics = {"avg_scores": {}, "std": {}}
        for metric_key in results_list[0].keys():
            aggregated_metrics["avg_scores"][metric_key] = np.mean(np.array([result[metric_key] for result in results_list]), axis=0)
            aggregated_metrics["std"][metric_key] = np.std(np.array([result[metric_key] for result in results_list]), axis=0)

        return aggregated_metrics

    def __repr__(self):
        avg_scores = self.report["avg_scores"]
        stds = self.report["std"]
        metris_string = [f'{metric_key + ":":50} {metric_scores[self.epoch]:10.4f} ({stds[metric_key][self.epoch]:.4f})'
                         for metric_key, metric_scores in avg_scores.items() if len(metric_scores.shape) == 1]

        report = \
            f"NESTED CV REPORT: {self.nested_cv_path} \n" + \
            f"Experiment ids: {[e.experiment_id for e in self.experiments]} \n" + \
            f"Hyper parameter combination ids: {[e.hyper_paramater_combination_id for e in self.experiments]} \n" + \
            "METRICS: \n" + \
            "\n".join(metris_string)
        return report


class NestedCVAnalyzer(EvaluationAnalyzerIF):
    def __init__(self, result_directory: str, scoring_fun: Callable[[ExperimentRepresentation], float]):
        self.result_directory = result_directory
        self.scoring_fun = scoring_fun

    @staticmethod
    def _load_experiments(result_directory: str) -> NestedCVRepresentation:
        nested_cv_representation = NestedCVRepresentation()
        metric_paths = list(sorted(glob.glob(os.path.join(result_directory, "**/*metrics.json"), recursive=True)))
        config_paths = list(sorted(glob.glob(os.path.join(result_directory, "**/*config.json"), recursive=True)))
        experiment_representations = [ExperimentRepresentation(config_path=config_path, metric_path=metric_path)
                                      for metric_path, config_path in tqdm.tqdm(zip(metric_paths, config_paths), desc="Loading experiments")]
        for e in experiment_representations:
            nested_cv_representation.add_experiment(e)
        return nested_cv_representation

    @staticmethod
    def _select_best_model_of_inner_fold(outer_fold_id: int, nested_cv_representation: NestedCVRepresentation,
                                         scoring_fun: Callable[[ExperimentRepresentation], float]) -> int:
        # inner fold id -> hpc_id -> Experiment
        experiments = nested_cv_representation.inner_experiments[outer_fold_id]
        # hpc_id -> [score_1, score_2]
        scores = defaultdict(list)
        for _, inner_fold in experiments.items():
            for hpc_id, experiment in inner_fold.items():
                scores[hpc_id].append(scoring_fun(experiment=experiment))

        # average the scores for hcp_id combination over the inner folds
        avg_scores = {key: np.mean(score) for key, score in scores.items()}
        keys = list(avg_scores.keys())
        values = list(avg_scores.values())
        best_hpc_id = keys[np.argmax(values)]
        return best_hpc_id

    def analyze(self, epoch: int) -> NestedCVReport:
        nested_cv_representation = self._load_experiments(self.result_directory)
        if not nested_cv_representation.verify():
            raise Exception("Corrputed Nested CV representation!")
        # we aggregate the scores on the inner folds and select the experiment in the outer fold
        # whose hyper parameter combinations had the best score on the inner folds
        best_full_experiments: List[ExperimentRepresentation] = []
        for outer_fold_id in nested_cv_representation.inner_experiments.keys():
            bets_hpc_id = self._select_best_model_of_inner_fold(outer_fold_id=outer_fold_id,
                                                                nested_cv_representation=nested_cv_representation,
                                                                scoring_fun=self.scoring_fun)
            best_experiment = nested_cv_representation.outer_experiments[outer_fold_id][bets_hpc_id]
            best_full_experiments.append(best_experiment)
        report = NestedCVReport(self.result_directory, best_full_experiments, epoch)
        return report


def scoring_fun_single_scalar_metric(metric_key: str, experiment: ExperimentRepresentation, epoch: int = -1) -> float:
    return experiment.results[metric_key][epoch]


# if __name__ == "__main__":
#     from functools import partial

#     epoch = -1
#     scoring_fun = partial(scoring_fun_single_scalar_metric, metric_key="train/auroc_macro", epoch=epoch)

#     result_directory = ""
#     analyzer = NestedCVAnalyzer(result_directory, scoring_fun=scoring_fun)
#     report = analyzer.analyze(epoch)
#     print(report)



from typing import Callable, Dict, Any, List
from functools import partial
from ml_gym.metrics.metrics import BinaryClasswiseExpectedCalibrationErrorMetric, PredictionMetric, BrierScoreMetric, \
    ClassSpecificExpectedCalibrationErrorMetric

from ml_gym.batching.batch import InferenceResultBatch


class MetricFactory:

    @staticmethod
    def get_sklearn_metric(metric_key: str, metric_fun: Callable,  params: Dict = None) -> Callable[[InferenceResultBatch], Any]:
        if params is None:
            params = {}
        return partial(PredictionMetric, identifier=metric_key, metric_fun=metric_fun, **params)

    @staticmethod
    def get_brier_score_metric_fun(tag: str,
                                   prediction_subscription_key: str,
                                   target_subscription_key: str):
        brier_score_fun = BrierScoreMetric(tag=tag,
                                           identifier="BRIER_SCORE",
                                           prediction_subscription_key=prediction_subscription_key,
                                           target_subscription_key=target_subscription_key)
        return brier_score_fun

    @staticmethod
    def get_expected_calibration_error_metric_fun(tag: str,
                                                  prediction_subscription_key: str,
                                                  target_subscription_key: str,
                                                  num_bins: int,
                                                  class_label: int,
                                                  sum_up_bins: bool):
        ece_score_fun = ClassSpecificExpectedCalibrationErrorMetric(tag=tag,
                                                                    identifier="EXPECTED_CALIBRATION_ERROR",
                                                                    prediction_subscription_key=prediction_subscription_key,
                                                                    target_subscription_key=target_subscription_key,
                                                                    num_bins=num_bins,
                                                                    class_label=class_label,
                                                                    sum_up_bins=sum_up_bins)
        return ece_score_fun

    @staticmethod
    def get_binary_classwise_expected_calibration_error_metric_fun(tag: str,
                                                                   target_subscription_key: str,
                                                                   prediction_subscription_key_0: str,
                                                                   prediction_subscription_key_1: str,
                                                                   num_bins: int,
                                                                   class_labels: List[int]):
        ece_score_fun = BinaryClasswiseExpectedCalibrationErrorMetric(tag=tag,
                                                                      identifier="BINARY_CLASSWISE_EXPECTED_CALIBRATION_ERROR",
                                                                      prediction_subscription_key_0=prediction_subscription_key_0,
                                                                      prediction_subscription_key_1=prediction_subscription_key_1,
                                                                      target_subscription_key=target_subscription_key,
                                                                      num_bins=num_bins,
                                                                      class_labels=class_labels)
        return ece_score_fun



from typing import Dict, Any, Type
from ml_gym.blueprints.blue_prints import create_blueprint
from ml_gym.gym.gym import Gym
from ml_gym.gym.jobs import AbstractGymJob
from ml_gym.validation.validator import ValidatorIF
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.util.grid_search import GridSearch


class GridSearchValidator(ValidatorIF):
    def __init__(self, grid_search_id: str, re_eval: bool = False):
        self.grid_search_id = grid_search_id
        self.re_eval = re_eval

    def run(self, blue_print_type: Type[BluePrint], gym: Gym, gs_config: Dict[str, Any],
            num_epochs: int, dashify_logging_path: str):
        run_id_to_config_dict = {run_id: config for run_id, config in enumerate(GridSearch.create_gs_from_config_dict(gs_config))}
        blueprints = []
        for config_id, experiment_config in run_id_to_config_dict.items():
            blueprint = create_blueprint(blue_print_class=blue_print_type,
                                         run_mode=AbstractGymJob.Mode.TRAIN if not self.re_eval else AbstractGymJob.Mode.EVAL,
                                         experiment_config=experiment_config,
                                         dashify_logging_path=dashify_logging_path,
                                         num_epochs=num_epochs,
                                         grid_search_id=self.grid_search_id,
                                         experiment_id=config_id)
            blueprints.append(blueprint)
        gym.add_blue_prints(blueprints)
        gym.run(parallel=True)



import torch
import os
from typing import List

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"


def get_devices(gpu_device_ids: List[int] = None) -> List[torch.device]:
    if torch.cuda.is_available():  # if we got some GPUs
        if gpu_device_ids is None:
            gpu_device_ids = list(range(torch.cuda.device_count()))
        return [torch.device(f"cuda:{device_id}") for device_id in gpu_device_ids]
    else:
        print("WARNING: No cuda devices available. Falling back to CPUs.")   # TODO Introduce proper logging...
        return [torch.device("cpu")]


if __name__ == "__main__":
    print(get_devices())



from typing import Dict, Any, Type
from ml_gym.validation.nested_cross_validation import NestedCV
from ml_gym.validation.gs_validator import GridSearchValidator
from ml_gym.blueprints.blue_prints import BluePrint


class ValidatorFactory:

    @staticmethod
    def get_nested_cv(gs_config: Dict[str, Any], cv_config: Dict[str, Any], grid_search_id: str,
                      blue_print_type: Type[BluePrint], re_eval: bool = False) -> NestedCV:
        iterator_key = cv_config["NestedCV"]["iterator_key"]
        split_key = cv_config["NestedCV"]["split_key"]
        component_names = [iterator_key]
        components = blue_print_type.construct_components(config=gs_config, component_names=component_names)
        iterator = components[iterator_key][split_key]
        return NestedCV(dataset_iterator=iterator, grid_search_id=grid_search_id, **cv_config["NestedCV"]["config"], re_eval=re_eval)

    @staticmethod
    def get_gs_validator(grid_search_id: str, re_eval: bool = False) -> GridSearchValidator:
        return GridSearchValidator(grid_search_id=grid_search_id, re_eval=re_eval)
